# frozen_string_literal: true

require 'csv'

# HTTP client library
module Faraday
  # Raised when any encoding error occurs
  EncodingError = Class.new(ClientError)

  # HTTP response
  class Response
    # Parses response bodies as CSV
    #
    # @example
    #
    #     require 'faraday/csv'
    #
    #     connection = Faraday.new('https://example.org/') do |faraday|
    #       faraday.response :csv, headers: :first_row
    #       faraday.adapter Faraday.default_adapter
    #     end
    #
    #     response = connection.get('/data.csv')
    #     response.body.first #=> <CSV::Row "foo":"bar">
    #
    # @api private
    class ParseCSV < Middleware
      # Creates a new instance of CSV parsing middleware
      #
      # @param app [#call, nil] Rack app
      # @param opts [Hash] CSV parsing options
      # @see CSV.new
      def initialize(app = nil, opts = {})
        @opts = opts

        super(app)
      end

      CSV_CONTENT_TYPE_RE = %r{
        \A # Start
        text/csv(?:;\s* # MIME type
          charset=(?<charset>[a-z0-9\-]+))? # Optional charset value
        \z # End
      }xi

      private_constant :CSV_CONTENT_TYPE_RE

      TARGET_ENCODING = Encoding::UTF_8
      private_constant :TARGET_ENCODING

      private

      attr_reader :opts

      def on_complete(env) # rubocop:disable MethodLength,AbcSize
        return unless env.parse_body?

        content_type = env.response_headers[:content_type]
        csv_content_type = content_type.match(CSV_CONTENT_TYPE_RE)

        return unless csv_content_type

        charset = csv_content_type[:charset]
        env.body.encode!(TARGET_ENCODING, charset) if charset

        env.body = CSV.parse(env.body, opts)
      rescue ::EncodingError => error
        raise Faraday::EncodingError.new(error, env.response)
      rescue CSV::MalformedCSVError => error
        raise Faraday::ParsingError.new(error, env.response)
      end
    end
  end
end
